import { LanguageSupport, LanguageDescription } from '@codemirror/language';
import { Tag } from "@codemirror/highlight";
import type { NodePropSource } from 'lezer-tree';
import type { Extension } from '@codemirror/state';
import type { IMonarchLanguage } from './monarch/monarchCommon';
/** The options / interface required to create a Monarch language. */
export interface MonarchLanguageDefinition {
    /** The name of the language. This is actually important for CodeMirror, so make sure it's correct. */
    name: string;
    /** The Monarch lexer that will be used to tokenize the language. */
    lexer: IMonarchLanguage;
    /** A list of `LanguageDescription` objects that will be used when the parser nests in a language. */
    nestLanguages?: LanguageDescription[];
    /** Configuration options for the parser, such as node props. */
    configure?: MonarchConfigure;
    /** A list of aliases for the name of the language. (e.g. 'go' -> ['golang']) */
    alias?: string[];
    /** A list of file extensions. (e.g. ['.ts']) */
    ext?: string[];
    /** The 'languageData' field for the language. CodeMirror plugins use this data to interact with the language. */
    languageData?: {
        [name: string]: any;
    };
    /** Extra extensions to be loaded. */
    extraExtensions?: Extension[];
}
/** An object containing the various utilites / loaders generated by a Monarch language. */
export interface MonarchLanguageData {
    /** Creates a `LanguageSupport` object that can be used like an ordinary language/extension. */
    load(): LanguageSupport;
    /** A list of new `Tag` objects generated automatically from the language definition. */
    tags: {
        [name: string]: Tag;
    };
    /** A `LanguageDescription` object, commonly used for nesting languages. */
    description: LanguageDescription;
}
/** Creates a new Monarch-based language.
 *  It returns an object containing:
 *
 *  | | |
 *  | :-- | :-- |
 *  | `load()` | a function used to load the language as an extension/`LanguageSupport` |
 *  | `description` | a `LanguageDescription` object, commonly used for nesting |
 *  | `tags` | a list of new `Tag` objects generated automatically from the language definition |
 */
export declare function createMonarchLanguage(opts: MonarchLanguageDefinition): MonarchLanguageData;
interface MonarchConfigure {
    props?: NodePropSource[];
}
export {};
