import { NodeSet, NodeType, Tree, stringInput } from 'lezer-tree';
import { Language, EditorParseContext, LanguageSupport, languageDataProp, defineLanguageFacet, LanguageDescription } from '@codemirror/language';
import { EditorState } from '@codemirror/state';
import { Tag, tags, styleTags } from "@codemirror/highlight";
import { compile } from './monarch/monarchCompile';
import { MonarchStack, stackIsEqual, tokenize } from './lexer';
/** Creates a new Monarch-based language.
 *  It returns an object containing:
 *
 *  | | |
 *  | :-- | :-- |
 *  | `load()` | a function used to load the language as an extension/`LanguageSupport` |
 *  | `description` | a `LanguageDescription` object, commonly used for nesting |
 *  | `tags` | a list of new `Tag` objects generated automatically from the language definition |
 */
export function createMonarchLanguage(opts) {
    // general chores before we can do stuff with the data
    const langDesc = Object.assign({ name: opts.name }, opts.alias ? { alias: opts.alias } : {}, opts.ext ? { extensions: opts.ext } : {});
    const langData = { ...langDesc, ...(opts.languageData ?? {}) };
    const lexer = compile(opts.lexer);
    // export tags
    const newTags = {};
    const unknownTags = Array.from(lexer.tokenTypes).filter(tag => !(tag in tags));
    unknownTags.forEach(tagName => { if (tagName)
        newTags[tagName] = Tag.define(); });
    const load = function () {
        const dataFacet = defineLanguageFacet(langData);
        const parser = createMonarchState({ lexer, configure: opts.configure ?? {}, tags: newTags, dataFacet, nestLanguages: opts.nestLanguages ?? [] });
        const startParse = (input, startPos, context) => {
            return monarchParse(parser, input, startPos, context);
        };
        const lang = new Language(dataFacet, { startParse });
        return new LanguageSupport(lang);
    };
    const description = LanguageDescription.of({ ...langDesc, load: async () => load() });
    return { load, tags: newTags, description };
}
// -- PARSER
// TODO: indent handling
// TODO: use monarch's brace handling automatically and get brace info into nodes
// TODO: allow 'emphasis.slash' where the '.slash' makes the 'emphasis' more specific, but uses the same scope
// TODO: use the tree fragments to get the exact edited text positions (relatively close, anyways)
// TODO: add action.transform
// (matches, stack) => FuzzyAction | FuzzyAction[] | null
// TODO: new caching strategy - store every line of the current document in a map
// that way it doesn't have to care about the specific line number
// after a new tree has been parsed, the map is cleared (or just replaced)
// that way there isn't any cache storage issues either, the map is just continously refreshed to the new state
// ? multi-line 'token' mode?
// ? inspect tokens widgets? pls cm6 add it urself
// https://gist.github.com/hyamamoto/fd435505d29ebfa3d9716fd2be8d42f0#gistcomment-2694461
function quickHash(s) {
    let h = 0;
    for (const c of s)
        h = Math.imul(31, h) + c.charCodeAt(0) | 0;
    return h;
}
/** Creates a `MonarchState`. I know, boring - see the `MonarchState` interface. */
function createMonarchState({ lexer, configure, tags: newTags, dataFacet, nestLanguages }) {
    const allTags = { ...tags, ...newTags };
    const lines = [];
    const nodeMap = new Map;
    let nodeTypes = [];
    let nodeSet = new NodeSet(nodeTypes);
    // this sets the language data facet on the top-most node of the language
    // the facet is how CodeMirror determines what language a region is using
    nodeMap.set('document', 0);
    nodeTypes.push(new NodeType("document", languageDataProp.set(Object.create(null), dataFacet), 0));
    // go through each token type and add it to our NodeTypes list
    lexer.tokenTypes.forEach((name) => {
        const id = nodeMap.size;
        nodeMap.set(name, id);
        let props = [];
        // check if the scope should be styled as a syntax highlighting tag (lower case)
        if (name[0].toUpperCase() !== name[0])
            props.push(styleTags({ [name + '/...']: allTags[name] ?? NodeType.none }));
        // push finished type
        nodeTypes.push(NodeType.define({ name, id, props }));
    });
    if ('props' in configure)
        nodeSet = nodeSet.extend(...configure.props);
    // since the state doesn't actually do anything by itself it's probably best as simply an object
    return {
        lexer,
        nodeMap,
        nodeTypes,
        nodeSet,
        lines,
        nestLanguages
    };
}
/** Compiles a mapped token from a `MonarchToken` and a mapping of scope names to `NodeType` IDs. */
function compileMappedToken(token, map) {
    let parserOpenAction;
    let parserCloseAction;
    if (token.parser) {
        if ('open' in token.parser)
            parserOpenAction = [map.get(token.parser.open), 0];
        if ('close' in token.parser)
            parserCloseAction = [map.get(token.parser.close), 0];
        if ('start' in token.parser)
            parserOpenAction = [map.get(token.parser.start), 1];
        if ('end' in token.parser)
            parserCloseAction = [map.get(token.parser.end), 1];
    }
    let tokenType = 0;
    if (token.type === '_NEST_')
        tokenType = -1;
    else if (token.type)
        tokenType = map.get(token.type);
    return [
        tokenType,
        token.start,
        token.end,
        parserOpenAction,
        parserCloseAction
    ];
}
class MonarchEmbeddedParser {
    state;
    range;
    lang = null;
    hash = 0;
    cache = Tree.empty;
    constructor(state, range) {
        this.state = state;
        this.range = range;
        if (state.nestLanguages.length) {
            this.lang = LanguageDescription.matchLanguageName(state.nestLanguages, range.lang);
            if (this.lang)
                this.lang.load();
        }
    }
    getParser() {
        if (this.lang?.support) {
            // parser loaded
            const host = this.lang.support.language.parser;
            return (input, startPos) => {
                const len = input.length;
                const hash = quickHash(input);
                if (hash === this.hash && this.cache.length > 0) {
                    // result already cached, return 'fake' `PartialParse`
                    this.hash = hash;
                    const cache = this.cache;
                    return { advance() { return cache; }, get pos() { return input.length; }, forceFinish() { return cache; } };
                }
                else {
                    // actual parser from language
                    this.hash = hash;
                    // isolated context shenanigans
                    // basically the nested parser is entirely unaware of the sorrounding document
                    const isolatedContext = new EditorParseContext(host, EditorState.create({ doc: input }), [], Tree.empty, { from: startPos, to: len }, []);
                    return host.startParse.bind(host)(stringInput(input), startPos, isolatedContext);
                }
            };
        }
        // parser not ready yet
        return (input, startPos, context) => EditorParseContext.skippingParser.startParse(stringInput(input), startPos, context);
    }
}
/** Represents a text line within the parser's cache. It is used to tokenize and compile strings as well. */
class MonarchLine {
    state;
    number;
    hash;
    length;
    startStack;
    endStack;
    tokens;
    embeds;
    lastOffset;
    lastBuffer;
    constructor(state, number, line) {
        this.state = state;
        this.number = number;
        this.reset(line);
        this.tokenize(line);
    }
    /** Resets the line using the given string (and hash, if already calculated) (does not tokenize). */
    reset(line, hash) {
        this.hash = hash ?? quickHash(line);
        this.length = line.length;
        this.lastOffset = -1;
        this.lastBuffer = [];
    }
    /** Returns the line prior to this one. */
    prev() { return this.state.lines[this.number - 1] ?? null; }
    /** Tokenizes the input string, and sets the line's state to reflect the result. */
    tokenize(line) {
        // get our starting state
        let stack;
        const lastLine = this.prev();
        if (lastLine)
            stack = new MonarchStack(lastLine.endStack);
        else
            stack = new MonarchStack([this.state.lexer.start ?? 'root']);
        this.startStack = stack.serialize();
        // tokenize
        const result = tokenize({ line, lexer: this.state.lexer, stack });
        this.tokens = result.tokens.map((token) => compileMappedToken(token, this.state.nodeMap));
        this.embeds = result.poppedEmbedded.map(range => new MonarchEmbeddedParser(this.state, range));
        this.endStack = stack.serialize();
    }
    /** Compiles the line's tokens into a linebuffer, with every position offset by the given `offset` value. */
    compile(offset) {
        if (this.lastOffset !== offset) {
            this.lastBuffer = this.tokens.map(token => [token[0], offset + token[1], offset + token[2], token[3], token[4]]);
            this.lastOffset = offset;
        }
        return this.lastBuffer;
    }
    /** Compares an input string to the current state.
     *  If the input is different from the state, the state will update to match the input.
     *  This function will return false if the lines are the same, and true if they differ.
     */
    eval(input) {
        let inHash;
        let invalid = false;
        const lastLine = this.prev();
        if (lastLine && !stackIsEqual(lastLine.endStack, this.startStack))
            invalid = true;
        else if (input.length !== this.length)
            invalid = true;
        else if ((inHash = quickHash(input)) !== this.hash)
            invalid = true;
        if (!invalid)
            return false;
        else {
            this.reset(input, inHash);
            this.tokenize(input);
            return true;
        }
    }
}
/** Returns a `PartialParse` compatible incremental parser using the given `MonarchState`. */
function monarchParse(state, input, start, context) {
    // set our viewport / start-end markers
    if (start < context.viewport.from)
        start = context.viewport.from;
    const viewportEndLine = context.state.doc.lineAt(context.viewport.to).number - 1;
    // next we want our list of lines from the document, but with them clipped to the input length
    const docLines = context.state.doc.slice(0, input.length).toJSON();
    // current line idx
    let idxLine = context.state.doc.lineAt(start).number - 1;
    // nesting, a line may have multiple nested langs so it needs to have an inline index
    let nesting = false;
    let nestingIdx = 0;
    let nestingParser = null;
    // we don't actually use a character pos, so we fake a function for it
    let early = false;
    const pos = () => {
        if (!early)
            return idxLine > docLines.length ? input.length : context.state.doc.line(idxLine).from;
        else
            return input.length;
    };
    // this function is called whenever codemirror wants to kill the parser and get the result
    // it's also called when the parser itself has nothing left to do (or leaves early)
    const getTree = () => {
        // go through the cache/lines and get all of our tokens in a list
        let offset = 0;
        const embeds = [];
        const tokens = state.lines.slice(0, docLines.length).flatMap(line => {
            let lineTokens = line.compile(offset);
            offset += line.length + 1;
            embeds.push(...line.embeds.map(embed => embed.cache));
            return lineTokens;
        });
        // here we're going to process `opens` and `closes` data and make an actually nesting tree
        let embedIdx = 0;
        let stack = [];
        const increment = () => stack.forEach(state => state[2]++);
        const buffer = [];
        for (const token of tokens) {
            // nesting lang handling
            if (token[0] === -1) {
                // undocumented features !!!!
                // passing a -1 to the size tells the tree builder to use the `reused` property
                // the ID of the token determines which slot in `reused` will take the place of this token
                buffer.push(embedIdx, token[1], token[2], -1);
                embedIdx++;
                increment();
            }
            else {
                // closing
                let closed = 0;
                if (token[4] && stack.length) {
                    const idx = stack.map(state => state[0]).lastIndexOf(token[4][0]);
                    if (idx !== -1) {
                        // cuts off anything past our closing stack element
                        stack = stack.slice(0, idx + 1);
                        // if we're inclusive of the end token we need to include it before we end the state
                        if (token[0] && token[4][1]) {
                            buffer.push(token[0], token[1], token[2], 4);
                            increment();
                        }
                        const state = stack.pop();
                        buffer.push(state[0], state[1], token[4][1] ? token[2] : token[1], (state[2] * 4) + 4);
                        increment();
                        closed = 1;
                    }
                }
                // token itself
                if (token[0] && !(token[4] && token[4][1])) {
                    buffer.push(token[0], token[1], token[2], 4);
                    increment();
                }
                // opening
                if (token[3]) {
                    stack.push([token[3][0], token[3][1] ? token[1] : token[2], token[0] && token[3][1] ? 1 : 0]);
                }
            }
        }
        // handle unfinished stack
        while (stack.length) {
            const state = stack.pop();
            buffer.push(state[0], state[1], pos(), (state[2] * 4) + 4);
            stack.forEach(state => state[2]++);
        }
        // handle embeddeds that never finish by replacing them with empty trees
        while (embeds.length <= embedIdx)
            embeds.push(Tree.empty);
        const tree = Tree.build({
            buffer: buffer,
            length: pos(),
            topID: 0,
            reused: embeds,
            nodeSet: state.nodeSet
        });
        return tree;
    };
    return {
        // this advances the parser one line and returns 'null' to signify it has done so
        // or, it may parse one line and then be 'complete', and return the parse tree
        // the advancing is controlled entirely by codemirror's scheduler
        advance() {
            const line = docLines[idxLine];
            let cachedLine = state.lines[idxLine];
            let lineUpdated = true;
            if (cachedLine)
                lineUpdated = cachedLine.eval(line);
            else {
                state.lines[idxLine] = new MonarchLine(state, idxLine, line);
                cachedLine = state.lines[idxLine];
            }
            // not parsing nested but needs to be
            if (!nesting && cachedLine.embeds[nestingIdx]) {
                nesting = true;
                const embed = cachedLine.embeds[nestingIdx];
                // get absolute start and end from the relative line offsets
                const start = context.state.doc.line(idxLine - embed.range.line + 1).from + embed.range.start;
                const end = context.state.doc.line(idxLine + 1).from + embed.range.end;
                nestingParser = embed.getParser()(input.read(start, end), 0, context);
            }
            // already parsing nested
            if (nesting && nestingParser) {
                const done = nestingParser.advance();
                if (done) {
                    nesting = false;
                    nestingParser = null;
                    cachedLine.embeds[nestingIdx].cache = done;
                    nestingIdx++;
                }
                else
                    return null;
            }
            // reset for next line
            nestingIdx = 0;
            idxLine++;
            // this basically catches when the user is just changing a single line
            // it will skip early if past the viewport and just use whatever is already in the cache until EOS
            if (docLines.length === state.lines.length && !lineUpdated && idxLine > viewportEndLine) {
                early = true;
                return getTree();
            }
            // EOS
            if (idxLine >= docLines.length)
                return getTree();
            // parsed one line, ready for next advance call
            else
                return null;
        },
        // these two functions complete the interface
        get pos() { return pos(); },
        forceFinish() { return getTree(); }
    };
}
//# sourceMappingURL=index.js.map