import type { IMonarchParserAction, ILexer } from './monarch/monarchCommon';
export declare type MonarchEmbeddedRange = {
    lang: string;
    line: number;
    start: number;
    end: number;
};
export declare class MonarchEmbeddedData {
    /** The language of the embedded data, e.g. 'javascript'. */
    lang: string;
    /** The starting line number of the embedded data, usually used as a _relative offset_ from the end line. */
    line: number;
    /** The stating offset from the beginning of the line. */
    start: number;
    constructor(
    /** The language of the embedded data, e.g. 'javascript'. */
    lang: string, 
    /** The starting line number of the embedded data, usually used as a _relative offset_ from the end line. */
    line: number, 
    /** The stating offset from the beginning of the line. */
    start: number);
    /** Serializes the embedded data into a string that can be attached to the front of a state `string[]` array. */
    serialize(): string;
    /** Returns the embedded data with a provided, final end position. */
    finalize(end: number): MonarchEmbeddedRange;
    /** Increments the line number by one, and returns the new number. */
    increment(): number;
    /** Returns a clone of the embedded data object. */
    clone(): MonarchEmbeddedData;
    /** Removes serialized embedded data from a state string. */
    static remove(state: string): string;
    /** Parses, and returns, an embedded data object from a state. */
    static deserialize(state: string): MonarchEmbeddedData | null;
}
/** The class used by Monarch to keep track of its internal state. */
export declare class MonarchStack {
    /** The internal `string[]` stack. */
    private stack;
    /** The embedded data, if present. */
    embedded: MonarchEmbeddedData | null;
    /** The top-most state of the stack. */
    get state(): string;
    /** The length (depth), or number of stack nodes, in the stack. */
    get depth(): number;
    /** The parent of the stack, i.e. the state of the stack if it were to be popped. */
    get parent(): MonarchStack;
    constructor(serializedStack: string[]);
    /** Push a new state to the top of the stack. */
    push(state: string): void;
    /** Switch to a new state, replacing the current one. */
    switchTo(state: string): void;
    /** Remove the top-most state of the stack. */
    pop(): string | undefined;
    /** Remove all states from the stack except the very first. */
    popall(): void;
    /** Returns a deep clone of the stack. */
    clone(): MonarchStack;
    /** Sets the embedded data. */
    setEmbedded(lang: string, line: number, start: number): void;
    /** Removes the embedded data. */
    endEmbedded(): MonarchEmbeddedData | null;
    /** Serializes the stack and embedded data into a list of strings. */
    serialize(): string[];
    /** Deserializes the stack and embedded data from a list of strings. */
    static deserialize(stack: string[]): {
        stack: string[];
        embedded: MonarchEmbeddedData | null;
    };
}
/** Compares two stacks and returns if they are equal.
 *  They can be pure `MonarchStack` objects or already serialized. */
export declare function stackIsEqual(stack1: string[] | MonarchStack, stack2: string[] | MonarchStack): boolean;
export interface MonarchToken {
    /** The type of the token. */
    type: string;
    /** The position of the start of the token, relative to the line it's on. */
    start: number;
    /** The position of the end of the token, relative to the line it's on. */
    end: number;
    /** Directs the parser to manipulate the syntax tree using the provided information. */
    parser?: IMonarchParserAction;
}
/** Determines whether the first token can be merged with the second. */
export declare function canContinueToken(lastToken?: MonarchToken, nextToken?: MonarchToken): boolean;
export interface TokenizeOpts {
    /** The line to be tokenized. */
    line: string;
    /** The (already compiled) Monarch lexer configuration. */
    lexer: ILexer;
    /** The `MonarchStack` that the tokenizer will begin with. */
    stack: MonarchStack;
    /** The offset from the start of the line that the tokenizer will begin from. */
    offset?: number;
}
/** Returns the tokens for the given line. Mutates the `opts.stack` `MonarchStack` object. */
export declare function tokenize(opts: TokenizeOpts): {
    stack: MonarchStack;
    tokens: MonarchToken[];
    poppedEmbedded: MonarchEmbeddedRange[];
};
